import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
import numpy as np
from clease.settings import ClusterExpansionSettings
from clease_gui import utils, BaseDashboard
import clease_gui as gui
from . import settings_buttons as swidgets

__all__ = ['SettingsDashboard']

logger = logging.getLogger(__name__)
gui.register_logger(logger)

DESC_WIDTH = '150px'
STYLE = {'description_width': DESC_WIDTH}


class SettingsDashboard(BaseDashboard):
    def initialize(self):
        self.cell_size = (3, 3, 3)
        self.out_size = widgets.Output()

        # Buttons related to the settings type,
        # CEBulk, CECrystal and CESlab

        # Output object for type buttons
        self.out_type = widgets.Output()
        self.out_help = widgets.Output()

        self.type_buttons_dct = {
            'CEBulk':
            swidgets.CEBulkButtons(self.app_data, output=self.out_type),
            'CECrystal':
            swidgets.CECrystalButtons(self.app_data, output=self.out_type),
            'CESlab':
            swidgets.CESlabButtons(self.app_data, output=self.out_type),
        }
        self.type_mode_b = widgets.Dropdown(
            options=[
                ('Bulk', 'CEBulk'),
                # Not yet supported
                # ('Crystal', 'CECrystal'),
                # ('Slab', 'CESlab'),
            ],
            description='Type:',
            layout={'width': 'max-content'},
            style=STYLE,
        )

        # Add observer to the type, so we know if we need to update the button collection
        def on_type_mode_change(change):
            if utils.is_value_change(change):
                self.update_type_mode()

        self.type_mode_b.observe(on_type_mode_change)

        self.size_mode_out = widgets.Output()
        self.size_mode_widgets = {
            'fixed':
            swidgets.FixedSizeMode(output=self.size_mode_out),
            'supercell_factor':
            swidgets.SupercellFactorSizeMode(output=self.size_mode_out),
        }
        self.size_mode_b = widgets.Dropdown(options=[('Fixed', 'fixed'),
                                                     ('Supercell Factor',
                                                      'supercell_factor')],
                                            value='supercell_factor',
                                            description='Size mode:',
                                            **self.DEFAULT_STYLE_KWARGS)
        self.size_mode_b.observe(self._on_size_mode_change)

    def update_type_mode(self):
        self.type_buttons.display_widgets()

    @property
    def type_buttons(self):
        mode = self.type_mode_b.value
        return self.type_buttons_dct[mode]

    def display(self):

        display(self.type_mode_b)
        display(self.out_type)
        self.update_type_mode()

        # Buttons for size mode
        with self.size_mode_out:
            self.active_size_mode_widget.display_widgets()
        display(self.size_mode_b, self.size_mode_out)

    @property
    def type_mode(self):
        return self.type_mode_b.value

    def get_type_kwargs(self):
        return self.type_buttons.value

    @property
    def active_size_mode(self):
        return self.size_mode_b.value

    @property
    def active_size_mode_widget(self):
        return self.size_mode_widgets[self.active_size_mode]

    def _on_size_mode_change(self, change):
        if utils.is_value_change(change):
            self._update_size_mode_widget()

    def _update_size_mode_widget(self):
        logger.debug('Updating active size mode widget')
        with self.size_mode_out:
            clear_output(wait=True)
            self.active_size_mode_widget.display_widgets()

    def get_settings_kwargs(self):
        size_kwargs = self.active_size_mode_widget.value
        kwargs = dict(
            type=self.type_mode,
            **size_kwargs,
            **self.get_type_kwargs(),
        )
        return kwargs

    def set_widgets_from_load(self,
                              settings: ClusterExpansionSettings) -> None:
        # Figure out what the builder was
        kwargs = settings.kwargs
        factory = kwargs.get('factory', None)
        if factory != 'CEBulk':
            logger.debug('Cannot load settings of factory type: %s', factory)
            return

        # Let's try and load things
        self._load_cebulk_settings(settings)

    def _load_cebulk_settings(self,
                              settings: ClusterExpansionSettings) -> None:
        mode = 'CEBulk'
        self.type_mode_b.value = mode
        self.update_type_mode()

        # Get some factory keys
        for key in ('crystalstructure', 'a', 'c', 'c_over_a'):
            value = settings.kwargs.get(key, None)
            if value is not None:
                self.type_buttons.set_widget_value(key, value)

        # Guess the size mode
        if settings.size is None:
            logger.debug('Detected supercell factor mode')
            # use supercell factor
            self.size_mode_b.value = 'supercell_factor'
            self._update_size_mode_widget()
            scf = settings.supercell_factor
            skew = settings.skew_threshold
            self.active_size_mode_widget.set_widget_value(
                'supercell_factor', scf)
            self.active_size_mode_widget.set_widget_value(
                'skew_threshold', skew)
        else:
            logger.debug('Detected fixed size mode')
            self.size_mode_b.value = 'fixed'
            # self._update_size_mode_widget()
            logger.debug('Getting size value')
            value = settings.size  # in 3x3 matrix
            # Use diagonal
            value = np.diag(value)
            value = ', '.join(map(str, value))
            logger.debug('Setting widget %s to value %s', key, value)
            self.active_size_mode_widget.set_widget_value('size', value)

        # Settings stored directly in the settings object
        all_keys = ('basis_func_type', 'db_name', 'max_cluster_size',
                    'max_cluster_dia', 'include_background_atoms')
        for key in all_keys:
            value = get_value_for_widget(settings, key)
            logger.debug('Setting %s to %s', key, value)
            self.type_buttons.set_widget_value(key, value)


def get_value_for_widget(settings, key):
    """Get a value from the settings, and sanitize it for settings
    a widget state """
    if key == 'basis_func_type':
        return settings.basis_func_type.name
    if key == 'db_name':
        value = settings.db_name
        return value
    if key == 'max_cluster_size':
        return settings.max_cluster_size
    if key == 'max_cluster_dia':
        mcd = settings.max_cluster_dia
        mcd = mcd[2:]  # Skip first two, they're 0 and 1 body clusters
        return mcd
    if key == 'include_background_atoms':
        return settings.include_background_atoms

    raise ValueError(f'Cannot deal with key: {key}')



from collections import namedtuple
from contextlib import contextmanager
import functools
import ipywidgets as widgets
from IPython.display import display

from clease_gui import BaseDashboard, Colors

__all__ = ['StatusBar', 'update_statusbar']

UpdatableWidget = namedtuple('UpdatableWidget', 'widget, update_function')

# _bool2text = {True: u'\N{GREEN CHECK MARK}', False: u'\N{CROSS MARK}'}
_bool2text = {True: u'\u2705', False: u'\N{CROSS MARK}'}


def update_statusbar(func):
    """Decorator to update statusbar after function completion, and mark
    status as busy from a dashboard object. """
    @functools.wraps(func)
    def _wrapper(self, *args, **kwargs):
        # statusbar = display_ui.main_dashboard.status_bar
        try:
            statusbar = self.app_data[self.KEYS.STATUS]['statusbar']
        except KeyError:
            # Something happened, we couldn't get the statusbar
            return func(self, *args, **kwargs)

        # Set statusbar as busy
        with statusbar.status_busy():
            return_val = func(self, *args, **kwargs)
        return return_val

    return _wrapper


class StatusBar(BaseDashboard):
    def __init__(self, *args, **kwargs):
        self._processes = 0
        self.all_widgets = []
        super().__init__(*args, **kwargs)
        # Register this statusbar in the app data
        # So others can register it as busy
        self.app_data[self.KEYS.STATUS] = {'statusbar': self}

    def initialize(self):
        # Create widgets
        self.status_widget = widgets.HTML(value=self.is_busy_str())
        self.register_status_widget(self.status_widget, self.is_busy_str)

        self.has_settings_wdgt = widgets.Label(value=self.has_settings())
        self.register_status_widget(self.has_settings_wdgt, self.has_settings)

        self.has_eci_wdgt = widgets.Label(value=self.has_eci())
        self.register_status_widget(self.has_eci_wdgt, self.has_eci)

        self.has_supercell_widget = widgets.Label(value=self.has_supercell())
        self.register_status_widget(self.has_supercell_widget,
                                    self.has_supercell)

    def display(self):
        hbox1 = widgets.VBox(
            children=[
                self.status_widget,
                self.has_settings_wdgt,
            ],
            # Add a bit of margin to the right
            layout=widgets.Layout(margin='0 50px 0 0'),
        )

        hbox2 = widgets.VBox(children=[
            self.has_eci_wdgt,
            self.has_supercell_widget,
        ])

        boxes = widgets.HBox(children=[hbox1, hbox2],
                             layout=widgets.Layout(border='solid 1px'))

        display(boxes)

    def register_status_widget(self, widget, func):
        updatable = UpdatableWidget(widget, func)
        self.all_widgets.append(updatable)

    @property
    def processes(self):
        return self._processes

    def has_supercell(self):
        val = self.KEYS.SUPERCELL in self.app_data
        return self._format_bool('Supercell available?', val)

    def is_busy(self):
        return self.processes > 0

    def is_busy_str(self):
        s = 'Status:   '
        if self.is_busy() > 0:
            # status = Colors.latex_red('Busy')
            status = Colors.html_color('Busy', 'red')
        else:
            status = Colors.html_color('Idle', 'green')
        return f'{s} {status}'

    def has_settings(self):
        # return 'settings' in self.app_data
        val = self.KEYS.SETTINGS in self.app_data
        return self._format_bool('Settings available?', val)

    def has_eci(self):
        val = self.KEYS.ECI in self.app_data
        return self._format_bool('ECI available?', val)

    def _format_bool(self, s, boolean):
        val = _bool2text[boolean]
        return f'{s}   {val}'

    def update(self):
        for updatable_widget in self.all_widgets:
            func = updatable_widget.update_function
            updatable_widget.widget.value = func()

    @contextmanager
    def status_busy(self):
        try:
            self._processes += 1
            self.update()
            yield
        finally:
            self._processes -= 1
            self.update()

    def __repr__(self):
        return f'{self.__class__.__name__}(busy={self.is_busy()}, processes={self.processes})'



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets

import clease
from clease.settings import CEBulk, CECrystal
from clease.tools import reconfigure
from clease_gui import register_logger, BaseDashboard, update_statusbar, utils
from . import ConcentrationDashboard, SettingsDashboard

__all__ = ['SettingsMakerDashboard']

logger = logging.getLogger(__name__)
register_logger(logger)


class SettingsMakerDashboard(BaseDashboard):
    """Have the concentration and settings dashboard under 1 tab"""
    def initialize(self):
        self.tab = None

        # dashboard = self  # Alias
        self.out_conc = widgets.Output()  # Concentration dashboard
        self.out_settings = widgets.Output()  # Settings dashboard

        self.make_settings_btn = widgets.Button(description='Make settings')
        self.make_settings_btn.on_click(self._on_make_settings_click)

        self.delete_settings_btn = widgets.Button(
            description='Delete settings', button_style='danger')
        self.delete_settings_btn.on_click(self._on_delete_settings_click)

        self.save_settings_btn = widgets.Button(description='Save settings')
        self.save_settings_fname = widgets.Text(description='Filename:',
                                                value='settings.json')
        self.save_settings_btn.on_click(self._on_save_settings_click)

        self.save_settings_box = widgets.HBox(children=[
            self.save_settings_btn,
            self.save_settings_fname,
        ])

        self.load_settings_btn = widgets.Button(description='Load settings')
        self.load_settings_fname = widgets.Text(description='Filename:',
                                                value='settings.json')
        self.load_settings_btn.on_click(self._on_load_settings_click)

        self.load_settings_box = widgets.HBox(children=[
            self.load_settings_btn,
            self.load_settings_fname,
        ])

        self._conc_dashboard = ConcentrationDashboard(self.app_data)
        with self.out_conc:
            clear_output()
            self._conc_dashboard.display()

        self._settings_dashboard = SettingsDashboard(self.app_data)
        with self.out_settings:
            clear_output()
            self._settings_dashboard.display()

        self.tab = widgets.Tab(children=[self.out_conc, self.out_settings])

        self.tab.set_title(0, 'Concentration')
        self.tab.set_title(1, 'Settings')

        self.reconfig_db_button = widgets.Button(description='Reconfigure DB')
        self.reconfig_db_button.on_click(self._on_reconfig_click)

    @update_statusbar
    @utils.disable_cls_widget('reconfig_db_button')
    def _on_reconfig_click(self, b):
        try:
            self.reconfig_db()
        except Exception as exc:
            self.log_error(logger, exc)

    def reconfig_db(self):
        settings = self.settings
        if settings is None:
            raise ValueError(
                'Settings has not been created yet, cannot reconfigure')
        # TODO: Account for CWD
        logger.info('Reconfiguring database: %s', settings.db_name)
        reconfigure(settings)
        logger.info('Reconfiguration complete.')

    def display(self):
        hbox_btn = widgets.HBox(
            children=[self.make_settings_btn, self.delete_settings_btn])
        display(
            hbox_btn,
            self.save_settings_box,
            self.load_settings_box,
            self.reconfig_db_button,
            self.tab,
        )

    def get_concentration(self):
        return self._conc_dashboard.get_concentration()

    @update_statusbar
    @utils.disable_cls_widget('make_settings_btn')
    def _on_make_settings_click(self, b):
        if self.settings is not None:
            logger.error(
                'Cannot make settings, a settings object already exists.')
            return
        logger.info('Making settings...')
        try:
            self.make_settings()
        except Exception as e:
            self.log_error(logger, 'Error when making settings: %s', e)
        else:
            logger.info('Done!')

    @update_statusbar
    @utils.disable_cls_widget('delete_settings_btn')
    def _on_delete_settings_click(self, b):
        try:
            self.app_data.pop(self.KEYS.SETTINGS)
            logger.info('Removed settings from app data')
        except KeyError:
            logger.info('No settings in app data, cannot delete.')

    def make_settings(self) -> None:
        """Combine the settings in the relevant dashboards, and create the settings"""
        conc = self.get_concentration()
        kwargs = self._settings_dashboard.get_settings_kwargs()
        skew_threshold = kwargs.pop('skew_threshold', None)
        settings_type = kwargs.pop('type')

        if settings_type == 'CEBulk':
            settings = CEBulk(conc, **kwargs)
        elif settings_type == 'CECrystal':
            settings = CECrystal(conc, **kwargs)
        elif settings_type == 'CESlab':
            # miller = settings_type.pop('miller')
            raise NotImplementedError('Not yet implemneted, yo!')
        else:
            raise ValueError(
                f'Something is wrong. Got unknown settings_type {settings_type}'
            )
        self.set_settings(settings)
        if skew_threshold is not None:
            self.settings.skew_threshold = skew_threshold

    @property
    def settings(self):
        """Get the settings from the app data"""
        return self.app_data.get(self.KEYS.SETTINGS, None)

    def set_settings(self, settings):
        """Insert settings into the app data"""
        self.app_data[self.KEYS.SETTINGS] = settings

    @property
    def cwd(self):
        return self.app_data[self.KEYS.CWD]

    def get_save_settings_fname(self):
        return self.cwd / self.save_settings_fname.value

    @update_statusbar
    @utils.disable_cls_widget('save_settings_btn')
    def _on_save_settings_click(self, b):
        try:
            fname = self.get_save_settings_fname()
            logger.info('Saving settings in file: %s', str(fname))
            self.save_settings(fname)
            logger.info('Save successful.')
        except Exception as exc:
            logger.error('Error during saving: %s', exc)

    def save_settings(self, fname):
        settings = self.settings
        if settings is None:
            raise RuntimeError('Settings has not been created yet.')
        settings.save(fname)

    def load_settings(self, fname) -> None:
        settings = clease.settings.settings_from_json(fname)
        self.set_settings(settings)
        self.set_settings_widget_states(settings)

    def get_load_settings_fname(self):
        return self.app_data[self.KEYS.CWD] / self.load_settings_fname.value

    @update_statusbar
    @utils.disable_cls_widget('load_settings_btn')
    def _on_load_settings_click(self, b):
        try:
            fname = self.get_load_settings_fname()
            logger.info('Loading settings from file: %s', str(fname))
            self.load_settings(fname)
            logger.info('Load successful.')
        except Exception as exc:
            self.log_error(logger, 'Error during loading: %s', exc)

    def set_settings_widget_states(self, settings):
        """Set the widget states from a new settings object"""
        conc = settings.concentration
        self._conc_dashboard.set_widgets_from_load(conc)
        self._settings_dashboard.set_widgets_from_load(settings)



# pylint: disable=undefined-variable,F401
from .colors import *
from .app_data import *
from .style_mixin import *
from .logging_widget import *
from .base_dashboard import *
from .status_bar import *
from .widget_collection import *
from .main import *
from . import utils
from . import settings_maker
from . import new_structures
from . import supercell

ADDITIONAL = ['utils', 'settings_maker', 'new_structures', 'supercell']

__all__ = (style_mixin.__all__ + base_dashboard.__all__ +
           logging_widget.__all__ + widget_collection.__all__ + main.__all__ +
           status_bar.__all__ + colors.__all__ + app_data.__all__) + ADDITIONAL



__all__ = ['Colors']


class Colors:
    RED = '\033[31m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    BOLD = '\033[1m'
    ITALIC = '\033[3m'
    RESET = '\033[0m'

    @staticmethod
    def html_color(text, color):
        return f'<font color="{color}">{text}</font>'



from .canonical_mc import *
from .plot_mc_page import *
from .mc_run_dashboard import *
from .mc_main import *

__all__ = canonical_mc.__all__ + mc_main.__all__ + plot_mc_page.__all__ + mc_run_dashboard.__all__



import ipywidgets as widgets

__all__ = ['WidgetStyleMixin']

STYLE = {'description_width': '150px'}
LAYOUT = {'width': 'max-content'}


class WidgetStyleMixin:
    DEFAULT_STYLE_KWARGS = {'style': STYLE, 'layout': LAYOUT}
    DEFAULT_BUTTON_LAYOUT = widgets.Layout(**LAYOUT)



from IPython.display import display
import ipywidgets as widgets
from clease_gui import BaseDashboard, utils

__all__ = ['GeneralNewStructSettingsDashboard']


class GeneralNewStructSettingsDashboard(BaseDashboard):
    # Automatically determine generation number by default?
    DEFAULT_AUTO_GEN = True

    def initialize(self):

        # Generation number box
        # None = automatically figure it out, add a checkbox for this
        self.gen_int_b = widgets.BoundedIntText(
            value=0,
            min=0,
            max=9999999,
            # We disable this box if we check figuring it out automatically
            disabled=self.DEFAULT_AUTO_GEN,
            description='Generation number:',
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.gen_auto_b = widgets.Checkbox(
            value=self.DEFAULT_AUTO_GEN,
            description='Automatically determine generation number?',
            **self.DEFAULT_STYLE_KWARGS)

        def on_gen_auto_change(change):
            """Disable the gen_int_b widget if we enable automatic
            generation ID determination"""
            if utils.is_value_change(change):
                self.gen_int_b.disabled = change['new']

        self.gen_auto_b.observe(on_gen_auto_change)

        self.gen_b = widgets.HBox(children=[self.gen_int_b, self.gen_auto_b])

        # Box for number of structures per generation
        self.struct_per_gen_b = widgets.BoundedIntText(
            value=5,
            min=1,
            max=9999999,
            description='No. struct per gen:',
            **self.DEFAULT_STYLE_KWARGS,
        )

    def display(self):
        display(self.gen_b, self.struct_per_gen_b)

    @property
    def generation_number(self):
        if self.gen_auto_b.value is True:
            # This uses the automatic determination
            return None
        return self.gen_int_b.value

    @property
    def struct_per_gen(self):
        return self.struct_per_gen_b.value



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
import clease
from clease_gui import register_logger, BaseDashboard, update_statusbar, utils
from .eci_fit import ECIFitDashboard
from .plots_dashboard import FitPlotDashboard

__all__ = ['FittingDashboard']

logger = logging.getLogger(__name__)
register_logger(logger, level=logging.DEBUG)


class FittingDashboard(BaseDashboard):
    def initialize(self):
        # Load ECI functionality
        self.load_eci_button = widgets.Button(description='Load ECI')
        self.load_eci_text = widgets.Text(description='Filename:',
                                          value='eci.json')
        self.load_eci_button.on_click(self._on_load_eci_click)

        # Save ECIs
        self.save_eci_button = widgets.Button(description='Save ECI')
        self.save_eci_text = widgets.Text(description='Filename:',
                                          value='eci.json')
        self.save_eci_button.on_click(self._on_save_eci_click)

        self.plotting_out = widgets.Output()
        self.fitting_eci_out = widgets.Output()
        eci_fit_dashboard = ECIFitDashboard(self.app_data)
        fit_plot_dashboard = FitPlotDashboard(self.app_data)

        with self.fitting_eci_out:
            clear_output()
            eci_fit_dashboard.display()

        with self.plotting_out:
            clear_output()
            fit_plot_dashboard.display()

        self.tab = widgets.Tab(children=[
            self.fitting_eci_out,
            self.plotting_out,
        ])

        self.tab.set_title(0, 'Fit ECI\'s')
        self.tab.set_title(1, 'Fit Plots')

    def display(self):
        load_hbox = widgets.HBox(
            children=[self.load_eci_button, self.load_eci_text])
        save_hbox = widgets.HBox(
            children=[self.save_eci_button, self.save_eci_text])
        display(load_hbox, save_hbox)
        display(self.tab)

    @update_statusbar
    @utils.disable_cls_widget('load_eci_button')
    def _on_load_eci_click(self, b):
        try:
            self._load_eci()
        except Exception as exc:
            logger.error(exc)

    def _load_eci(self):
        fname = self.load_eci_text.value
        if fname == '':
            raise ValueError('No filename supplied')
        fname = self.app_data[self.KEYS.CWD] / fname
        logger.info('Loading ECI from file: %s', fname)
        with fname.open() as file:
            eci = clease.jsonio.read_json(file)
        self.set_eci(eci)
        logger.info('Load successful.')

    @property
    def eci(self):
        try:
            return self.app_data[self.KEYS.ECI]
        except KeyError:
            raise ValueError(
                'No ECI\'s present. Run fit first, or load ECI from file.')

    def set_eci(self, value):
        self.app_data[self.KEYS.ECI] = value

    @update_statusbar
    @utils.disable_cls_widget('save_eci_button')
    def _on_save_eci_click(self, b):
        try:
            self._save_eci()
        except Exception as exc:
            logger.error('Error during saving: %s', exc)

    def _save_eci(self):
        eci = self.eci
        fname = self.save_eci_text.value
        if fname == '':
            raise ValueError('No filename supplied.')
        fname = self.app_data[self.KEYS.CWD] / fname
        logger.info('Saving ECI to file: %s', fname)
        with fname.open('w') as file:
            # Use CLEASE jsonio module to encode the ECI dict
            # since it contains NumPy arrays.
            clease.jsonio.write_json(file, eci)
        logger.info('Save successful.')



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
import ase.visualize
import pandas as pd

from clease_gui import BaseDashboard, update_statusbar
import clease_gui as gui

__all__ = ['DBViewerDashboard']

logger = logging.getLogger(__name__)
gui.register_logger(logger)


class DBViewerDashboard(BaseDashboard):
    def initialize(self):

        self.db_select_widget = widgets.Text(description='DB name:', value='')
        db_name_help = widgets.Label(('Leave empty to use database specified '
                                      'in the settings object.'))
        self.db_select_box = widgets.HBox(
            children=[self.db_select_widget, db_name_help])

        self.selection_parameters_widget = widgets.Text(
            description='Selection:', value='')

        selection_help_label = widgets.Label(
            ('For more information on querying, see '
             'https://wiki.fysik.dtu.dk/ase/ase/db/db.html#querying'))
        self.selection_box = widgets.HBox(
            children=[self.selection_parameters_widget, selection_help_label])

        self.draw_selection_button = widgets.Button(
            description='Open in ASE GUI')
        self.draw_selection_button.on_click(self._on_draw_selection_click)

        self.print_db_button = widgets.Button(description='Show DB content')
        self.print_db_button.on_click(self._on_print_db)

        # Output for displaying the ASE database as a pandas DF
        self.db_content_out = widgets.Output(layout=dict(
            overflow='auto',
            width='100%',
            height='350px',
        ))

    @gui.update_statusbar
    @gui.utils.disable_cls_widget('print_db_button')
    def _on_print_db(self, b):
        try:
            self._print_db()
        except Exception as exc:
            logger.error(exc)

    def _print_db(self):
        db_name = str(self.db_name)
        selection = self.selection_parameters_widget.value
        con = self.get_connection()
        out = self.db_content_out
        logger.info('Querying database %s', db_name)

        rows = con.select(selection)
        df = ase_db_rows_to_df(rows)
        with out:
            clear_output(wait=True)
            with pd.option_context('display.max_rows', None,
                                   'display.max_columns', None):
                # Temporarily set display context to all of the rows
                # and columns we have.
                # We don't wanna truncate here.
                display(df)

    def display(self):
        buttons = widgets.HBox(children=[
            self.print_db_button,
            self.draw_selection_button,
        ])
        display(
            self.db_select_box,
            self.selection_box,
            buttons,
            self.db_content_out,
        )

    @property
    def db_name(self):
        value = self.db_select_widget.value
        if value == '':
            # Use DB name from settings
            return self.get_db_name()
        return self.app_data[self.KEYS.CWD] / value

    @property
    def settings(self):
        return self.app_data[self.KEYS.SETTINGS]

    def get_connection(self):
        return ase.db.connect(self.db_name)

    def get_selection_parameters(self):
        value = self.selection_parameters_widget.value
        if value == '':
            return None
        return value

    def get_selection(self):
        selection = self.get_selection_parameters()
        con = self.get_connection()

        return [row.toatoms() for row in con.select(selection=selection)]

    @update_statusbar
    def _on_draw_selection_click(self, b):
        self._draw_selection()

    def _draw_selection(self):
        images = self.get_selection()
        ase.visualize.view(images)


def ase_db_rows_to_df(rows) -> pd.DataFrame:
    """Convert an ASE database into a pandas data frame"""
    rows = list(rows)
    base_keys = [
        'id', 'formula', 'calculator', 'energy', 'natoms', 'pbc', 'volume',
        'charge', 'mass'
    ]

    # Grab all extra keys in the key_value_pairs
    extra_keys = []
    for row in rows:
        keys = row.key_value_pairs.keys()
        for key in keys:
            if key not in extra_keys:
                extra_keys.append(key)

    all_keys = base_keys + extra_keys
    # Now populate the data
    rows_list = []
    for row in rows:
        data = {}
        for key in base_keys:
            try:
                value = getattr(row, key)
            except AttributeError:
                value = None
            data[key] = _formatter(key, value)
        kvp_dct = row.key_value_pairs
        for key in extra_keys:
            value = kvp_dct.get(key, None)
            data[key] = _formatter(key, value)
        rows_list.append(data)
    return pd.DataFrame(rows_list, columns=all_keys)


def _formatter(key, value):
    """Helper function to format the value, possibly
    on the basis of the name of the key"""
    if value is None:
        value = '-'
    if key == 'pbc':
        # Convert [True, True, True] => 'TTT'
        # or [True, False, False] => 'TFF'
        value = ''.join([str(v)[0] for v in value])

    return value



import os
import ipywidgets as widgets
from IPython.display import display
import logging

__all__ = [
    'default_handler',
    'register_logger',
    'OutputWidgetHandler',
    'initialize_clease_gui_logging',
]


class OutputWidgetHandler(logging.Handler):
    """ Custom logging handler sending logs to an output widget.
    Class inspired by
    https://ipywidgets.readthedocs.io/en/stable/examples/Output%20Widget.html
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # layout = widgets.Layout(width='100%',
        # height='160px',
        # bor)
        layout = {
            'width': '100%',
            'height': '160px',
            'border': '1px solid black',
            'overflow': 'auto',
        }
        self.out = widgets.Output(layout=layout)

    def emit(self, record):
        """ Overload of logging.Handler method """
        formatted_record = self.format(record)
        new_output = {
            'name': 'stdout',
            'output_type': 'stream',
            'text': formatted_record + os.linesep
        }
        self.out.outputs = (new_output, ) + self.out.outputs

    def show_logs(self):
        """ Show the logs """
        display(self.out)

    def clear_logs(self):
        """ Clear the current logs """
        self.out.clear_output()


default_handler = OutputWidgetHandler()
default_handler.setFormatter(
    logging.Formatter('%(asctime)s  - [%(levelname)s] %(message)s'))
registered_loggers = []


def register_logger(logger, level=logging.INFO):
    """Register a logger to the default output widget handler.
    Anything emitted by this log will be shown in that widget"""
    global default_handler, registered_loggers
    registered_loggers.append(logger)
    logger.addHandler(default_handler)
    logger.setLevel(level)


def _register_warning_logger():
    """Inspired by
    https://stackoverflow.com/questions/38531786/capturewarnings-set-to-true-doesnt-capture-warnings
    Warning: this changes global variable in logging.
    """
    logging.captureWarnings(True)
    warnings_logger = logging.getLogger("py.warnings")
    register_logger(warnings_logger)


def _register_clease_logger(level=logging.INFO):
    clease_logger = logging.getLogger('clease')
    register_logger(clease_logger, level=level)


def set_all_levels(level) -> None:
    for logger in registered_loggers:
        logger.setLevel(level)


def initialize_clease_gui_logging():
    # _register_clease_logger()
    _register_warning_logger()



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
from ase.data import atomic_numbers
from clease.settings import Concentration

from clease_gui import BaseDashboard, update_statusbar
import clease_gui as gui

__all__ = ['ConcentrationDashboard']

logger = logging.getLogger(__name__)
gui.register_logger(logger)


class ConcentrationDashboard(BaseDashboard):
    def initialize(self):
        self.basis_elements = []
        self.basis_text_box = widgets.Text(description='Basis:')
        self.out_basis = widgets.Output(layout=widgets.Layout(
            border='solid 1px'))

        self.add_basis_b = widgets.Button(description='Add basis')
        self.remove_basis_b = widgets.Button(description='Remove last basis')
        self.clear_basis_b = widgets.Button(description='Clear basis')

        self.add_basis_b.on_click(self._on_add_basis)
        self.remove_basis_b.on_click(self._on_remove_last_basis)
        self.clear_basis_b.on_click(self._on_clear_basis)

    def display(self):
        print('Add or remove basis elements to the concentration')
        self.update_basis_out()

        display(
            self.basis_text_box,
            self.add_basis_b,
            self.remove_basis_b,
            self.clear_basis_b,
            self.out_basis,
        )

    def update_basis_out(self):
        with self.out_basis:
            clear_output()
            print('Current basis:', self.basis_elements)

    def add_basis(self, basis):
        for element in basis:
            if element not in atomic_numbers:
                self.bad_basis_element(element)
                return
        self.basis_elements.append(basis)
        self.update_basis_out()

    def bad_basis_element(self, element):
        self.update_basis_out()
        with self.out_basis:
            print('Bad element found:', element)

    @update_statusbar
    def _on_add_basis(self, b):
        value = self.basis_text_box.value

        # Replace newlines with , for string manupulation
        DELIM = ' '
        value = value.replace('\n', DELIM)
        value = value.replace(',', DELIM)
        value = value.replace(';', DELIM)

        if not value:
            # Button was clicked without adding any elements
            return

        new_basis = [
            val.strip() for val in value.strip().split(DELIM) if val.strip()
        ]
        self.add_basis(new_basis)

    def remove_last_basis(self, update=True):
        try:
            self.basis_elements.pop()
        except IndexError:
            # Empty
            pass
        if update:
            self.update_basis_out()

    @update_statusbar
    def _on_remove_last_basis(self, b):
        self.remove_last_basis()

    @update_statusbar
    def _on_clear_basis(self, b):
        tot = len(self.basis_elements)
        for _ in range(tot):
            self.remove_last_basis(update=False)
        self.update_basis_out()

    def get_concentration(self):
        return Concentration(basis_elements=self.basis_elements)

    def set_widgets_from_load(self, conc: Concentration) -> None:
        logger.debug('Updating conc widgets')
        self.basis_elements = conc.basis_elements
        self.update_basis_out()



import logging
import uuid
import threading
from IPython.display import display, clear_output
import ipywidgets as widgets
import numpy as np
import clease
from clease.montecarlo import constraints
import clease_gui as gui

__all__ = ['CanonicalMC']

logger = logging.getLogger(__name__)
gui.register_logger(logger)

# Register the clease MC logger specifically.
# Allows to follow the internal messages from the MC object.
# clease_mc_logger = logging.getLogger('clease.montecarlo.montecarlo')
# gui.register_logger(clease_mc_logger)


class CanonicalMC(gui.BaseDashboard):
    def __init__(self, *args, **kwargs):
        self._mc = None  # MC process
        super().__init__(*args, **kwargs)

    def initialize(self):
        # Temperature widgets

        self.temperature_mode_widget = widgets.Dropdown(
            description='MC mode:',
            options=[('Fixed Temperature', 'temp_fixed'),
                     ('Simulated annealing (linear spacing)',
                      'temp_range_linspace'),
                     ('Simulated annealing (logarithmic spacing)',
                      'temp_range_logspace')],
            value='temp_range_logspace',
            **self.DEFAULT_STYLE_KWARGS,
        )

        temp_range_linear = make_spacing_boxes(
            style_kwargs=self.DEFAULT_STYLE_KWARGS, )

        temp_range_log = make_spacing_boxes(
            style_kwargs=self.DEFAULT_STYLE_KWARGS, )

        self.temp_widgets = {
            'temp_fixed':
            widgets.BoundedFloatText(
                value=300,
                min=0,
                max=9999999,
                description='Temperature:',
                **self.DEFAULT_STYLE_KWARGS,
            ),
            'temp_range_linspace':
            temp_range_linear,
            'temp_range_logspace':
            temp_range_log,
        }
        self.temp_mode_output = widgets.Output()
        self._update_temp_mode()

        def _on_temp_mode_change(change):
            """Update the steps_output display whenever we change the dropdown"""
            if gui.utils.is_value_change(change):
                self._update_temp_mode()

        self.temperature_mode_widget.observe(_on_temp_mode_change)

        # MC logging frequency
        self.update_freq_widget = widgets.BoundedIntText(
            value=30,
            min=1,
            max=9999999,
            description='Update freq. (sec):',
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.update_freq_help_msg = widgets.Label(
            value='Frequency of the logging from the MC run (in seconds).')

        self.step_mode_help = widgets.Label(
            '1 sweep runs equals <number of atoms in supercell> steps.')

        # Number of steps

        self.step_widgets = {}
        sweeps_widget = widgets.BoundedIntText(
            description='Number of sweeps:',
            value=1,
            min=1,
            max=99999999,
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.step_widgets['sweeps'] = sweeps_widget

        steps_widget = widgets.BoundedIntText(
            description='Number of steps:',
            value=100,
            min=1,
            max=999999999999999,
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.step_widgets['steps'] = steps_widget

        self.steps_output = widgets.Output()
        # Dropdown for selecting the step mode
        self.step_mode_widget = widgets.Dropdown(
            options=[
                ('Sweeps', 'sweeps'),
                ('Steps', 'steps'),
            ],
            description='Step mode:',
            value='sweeps',
            **self.DEFAULT_STYLE_KWARGS,
        )
        self._update_step_mode_output()  # update the steps_output

        def _on_step_mode_change(change):
            """Update the steps_output display whenever we change the dropdown"""
            if gui.utils.is_value_change(change):
                self._update_step_mode_output()

        self.step_mode_widget.observe(_on_step_mode_change)

        self.run_mc_button = widgets.Button(description='Run MC',
                                            button_style='primary')
        self.run_mc_button.on_click(self._on_run_mc_click)

        self.update_atoms_after_run_widget = widgets.Checkbox(
            description='Update supercell after MC run?', value=True)

        self.clear_mc_app_data_button = widgets.Button(
            description='Clear MC data')
        self.clear_mc_app_data_button.on_click(self._on_clear_mc_data)

        # Button for stopping the MC process
        self.stop_mc_button = widgets.Button(description='Abort MC',
                                             button_style='warning')
        self.stop_mc_button.on_click(self._on_stop_mc_click)

        self.mc_progress = widgets.IntProgress()
        self.mc_progress_label = widgets.Label()

        self.constrain_sublattices_wdgt = widgets.Checkbox(
            value=True,
            description='Constrain swaps by sublattice?',
            **self.DEFAULT_STYLE_KWARGS)

    def attach_constraints(
            self, mc: clease.montecarlo.Montecarlo,
            settings: clease.settings.ClusterExpansionSettings) -> None:
        if self.constrain_sublattices_wdgt.value:
            atoms = mc.atoms
            cnst = constraints.ConstrainSwapByBasis(atoms,
                                                    settings.index_by_basis)
            mc.generator.add_constraint(cnst)

    @property
    def update_after_run(self) -> bool:
        """Do we update the supercell in app_data after
        a completed MC run?"""
        return self.update_atoms_after_run_widget.value

    def display(self):
        frq_box = widgets.HBox(children=[
            self.update_freq_widget,
            self.update_freq_help_msg,
        ])
        step_mode_box = widgets.HBox(children=[
            self.step_mode_widget,
            self.step_mode_help,
        ])

        # place update atoms checkbox next to run MC button
        run_mc_box = widgets.HBox(
            children=[
                self.run_mc_button,
                self.stop_mc_button,
                self.update_atoms_after_run_widget,
            ],
            layout=dict(justify_content='space-between'),
        )

        display(step_mode_box, self.steps_output)
        display(self.temperature_mode_widget, self.temp_mode_output)
        display(frq_box, self.constrain_sublattices_wdgt)
        display(self.clear_mc_app_data_button)

        display(run_mc_box)
        mc_progress_box = widgets.HBox(children=[
            self.mc_progress,
            self.mc_progress_label,
        ])
        display(mc_progress_box)

    @property
    def active_temp_mode(self):
        return self.temperature_mode_widget.value

    @property
    def active_temp_mode_widget(self):
        return self.temp_widgets[self.active_temp_mode]

    def _update_temp_mode(self):
        with self.temp_mode_output:
            clear_output(wait=True)
            display(self.active_temp_mode_widget)

    def get_temp(self):
        mode = self.active_temp_mode
        widget = self.active_temp_mode_widget
        if mode == 'temp_fixed':
            value = widget.value
            return np.array([value])
        elif mode == 'temp_range_linspace':
            t_max = widget.children[0].value
            t_min = widget.children[1].value
            t_steps = widget.children[2].value
            return np.linspace(t_max, t_min, t_steps)
        elif mode == 'temp_range_logspace':
            t_max = widget.children[0].value
            t_min = widget.children[1].value
            t_steps = widget.children[2].value
            # Logspace is in base 10 by default
            return np.logspace(np.log10(t_max), np.log10(t_min), t_steps)
        raise ValueError(f'Unknown mode: {mode}')

    def _get_key(self, key):
        try:
            return self.app_data[key]
        except KeyError:
            raise KeyError(
                f'No key named {key} in app data. Must be loaded/created first.'
            )

    @property
    def active_step_mode(self):
        return self.step_mode_widget.value

    @property
    def active_step_mode_widget(self):
        mode = self.active_step_mode
        return self.step_widgets[mode]

    def _update_step_mode_output(self):
        with self.steps_output:
            clear_output(wait=True)
            display(self.active_step_mode_widget)

    def get_number_of_steps(self):
        mode = self.active_step_mode
        value = self.active_step_mode_widget.value
        if mode == 'sweeps':
            return value * len(self.atoms)
        assert mode == 'steps'
        return value

    @property
    def atoms(self):
        return self._get_key(self.KEYS.SUPERCELL)

    @property
    def eci(self):
        return self._get_key(self.KEYS.ECI)

    @property
    def settings(self):
        return self._get_key(self.KEYS.SETTINGS)

    @property
    def update_freq(self):
        return self.update_freq_widget.value

    def set_supercell(self, supercell):
        self.app_data[self.KEYS.SUPERCELL] = supercell

    def _make_mc_object(self, atoms, temp=200):
        eci = self.eci
        settings = self.settings

        # attach a CLEASE calculator
        atoms = clease.calculator.attach_calculator(settings, atoms, eci=eci)

        mc = clease.montecarlo.Montecarlo(atoms, temp)
        mc.status_every_sec = self.update_freq

        self.attach_constraints(mc, settings)

        return mc

    def _on_clear_mc_data(self, b):
        """Remove all data from previous MC runs"""
        try:
            del self.app_data[self.KEYS.CANONICAL_MC_DATA]
            logger.info('MC data has been cleared')
        except KeyError:
            logger.info('No MC data to clear.')
            # There was no data
            pass

    def set_mc_data(self, data):
        self.app_data[self.KEYS.CANONICAL_MC_DATA] = data

    @property
    def mc_data(self) -> list:
        return self._get_key(self.KEYS.CANONICAL_MC_DATA)

    def get_mc_data(self) -> list:
        """Return existing app data if it exists, otherwise return a new
        MC data list."""
        try:
            return self.mc_data
        except KeyError:
            # No data exists, create a new data dictionary
            # and insert it
            data = []
            self.set_mc_data(data)
            return self.mc_data

    def _on_stop_mc_click(self, b):
        """Set a termination flag in the currently active MC object,
        if there is one"""
        if self._mc is None:
            return
        self._mc.quit = True

    def _check_required_keys(self) -> None:
        """Check that relevant keys  for a MC run is present.
        Raises an error if that's not the case.
        """
        missing_keys = []
        for key in (self.KEYS.SUPERCELL, self.KEYS.SETTINGS, self.KEYS.ECI):
            if key not in self.app_data:
                missing_keys.append(key)
        if len(missing_keys) > 0:
            missing_keys = ', '.join(missing_keys)
            raise KeyError(
                f'Missing the following in app data: "{missing_keys}".')

    def _run_mc_in_thread(self) -> None:
        """Start the `_run_mc` method in a separate thread,
        while also disabling the run mc button for the duration
        of the run
        """
        @gui.update_statusbar
        def runner(self):
            """Helper function to be executed by the thread"""
            # Disable the run MC button while an MC thread is running
            try:
                with gui.utils.disable_widget_context(self.run_mc_button):
                    self._run_mc()
            except Exception as exc:
                self.log_error(logger, exc)

        # First we check if we have all necessary quantities, before
        # launching the thread
        self._check_required_keys()

        # The update_statusbar decorator assumes "self" is the first argument
        threading.Thread(target=runner, args=(self, ), daemon=True).start()

    def _run_mc(self):
        """Execute the MC run at the currnet settings."""
        # Some initialization
        temperatures = self.get_temp()
        n_temp = len(temperatures)

        mc_app_data = self.get_mc_data()
        data = {}
        mc_app_data.append(data)

        # Initial atoms object, current working atoms
        atoms = self.atoms

        def check_abort(mc):
            """Helper function to raise if we abort the MC.
            Raises an error, to break the thread which is running
            it"""
            if mc.quit:
                val = self.mc_progress_label.value
                self.mc_progress_label.value = val + '. MC aborted!'
                raise RuntimeError('MC was aborted.')

        mc = self._make_mc_object(atoms)
        # Add emin observer
        emin_obs = clease.montecarlo.observers.LowestEnergyStructure(mc.atoms)
        mc.attach(emin_obs)
        self._mc = mc
        # Initialize the progress bar
        self.mc_progress.value = 0
        self.mc_progress.max = n_temp

        try:
            for ii, temp in enumerate(temperatures):
                # Update progress bar
                self.mc_progress.value += 1
                prog = (ii + 1) / n_temp * 100  # Progress in percentage
                self.mc_progress_label.value = (
                    f'{prog:.1f}%, {ii+1}/{n_temp}. '
                    f'Current temp: {temp:.1f} K')

                # Initialize MC things
                check_abort(mc)
                mc.T = temp
                steps = self.get_number_of_steps()
                initial_energy = mc.atoms.get_potential_energy()

                # some debugging logging
                logger.debug(
                    'Starting MC run at T = %.3f K. Running a total of %d MC steps.',
                    temp, steps)
                # Simmulated annealing
                logger.debug(
                    'Annealing: Temperature step %d of %d. Current temperature: %.3f K.',
                    ii + 1,
                    n_temp,
                    temp,
                )

                mc.run(steps)  # Execute the MC
                check_abort(mc)  # Check if we exited because of an abort
                logger.debug('MC done after %d MC steps.', steps)
                logger.debug('Energy before MC: %.3f eV.', initial_energy)
                logger.debug('Lowest energy achieved during run: %.3f eV.',
                             emin_obs.lowest_energy)
                self._update_mc_data(mc, emin_obs, data)
                # Update the atoms object to the current emin, and continue from that
                atoms = emin_obs.emin_atoms

            if self.update_after_run:
                # "atoms" is now the emin atoms
                self.set_supercell(atoms)
                logger.info(
                    'Updated supercell to minimum energy configuration')

            logger.info('MC run complete.')
        finally:
            self._mc = None

    @gui.update_statusbar
    def _on_run_mc_click(self, b):
        try:
            self._run_mc_in_thread()
        except Exception as exc:
            logger.error(exc)

    def _make_mc_data(self, mc: clease.montecarlo.Montecarlo,
                      emin_obs) -> dict:
        """Create a dictionary with the relevant data from the MC run.
        We pass in the relevant emin observer as well, because we always want this data."""
        lowest_energy = emin_obs.lowest_energy
        temp = mc.T
        accept_rate = mc.num_accepted / mc.current_step
        mean_energy = mc.mean_energy.mean

        data = dict(
            emin=lowest_energy,
            mean_energy=mean_energy,
            temperature=temp,
            accept_rate=accept_rate,
        )
        return data

    def _update_mc_meta(self, mc, current_data):
        """Add metadata to the current data results"""
        atoms = mc.atoms
        current_meta = current_data.get('meta', {})
        meta = dict(
            natoms=len(atoms),
            formula=atoms.get_chemical_formula(),
            temp_mode=self.active_temp_mode,
        )
        current_meta.update(meta)
        if 'run_id' not in current_meta:
            current_meta['run_id'] = str(uuid.uuid4())
        current_data['meta'] = current_meta

    def _update_mc_data(self, mc: clease.montecarlo.Montecarlo, emin_obs,
                        current_data) -> None:
        """Update the current data with the new data"""
        new_data = self._make_mc_data(mc, emin_obs)
        self._update_mc_meta(mc, current_data)

        for key, value in new_data.items():
            if key not in current_data:
                # Add an empty list if this key has not been added yet
                # (emulating a defaultdict)
                current_data[key] = []
            current_data[key].append(value)


def make_spacing_boxes(min=300, max=30000, steps=25, style_kwargs=None):
    style_kwargs = style_kwargs or {}
    temp_range_min = widgets.BoundedFloatText(
        description='Temp (min):',
        min=1,
        max=99999999,
        value=min,
        **style_kwargs,
    )
    temp_range_max = widgets.BoundedFloatText(
        description='Temp. (max):',
        min=1,
        max=99999999,
        value=max,
        **style_kwargs,
    )
    temp_range_steps = widgets.BoundedIntText(
        description='Number of temp. steps:',
        min=1,
        max=99999999,
        value=steps,
        **style_kwargs,
    )

    temp_range_box = widgets.VBox(
        children=[temp_range_max, temp_range_min, temp_range_steps],
        layout=dict(justify_content='space-between'),
    )
    return temp_range_box



import logging
from abc import ABC
from IPython.display import display, clear_output
import ipywidgets as widgets
import numpy as np
import clease_gui as gui
from clease_gui.widget_collection import WidgetCollection, SingleWidget

logger = logging.getLogger(__name__)
gui.register_logger(logger)


class MaxClusterDiaWidget(SingleWidget):
    """Special widget for max cluster diameters, it needs to update itself,
    and recreate the VBox every time it needs to update."""
    DEFAULT_MCD_VALUE = 5.0

    def __init__(self, name, num_boxes, priority=0):
        self.mcd_boxes = []
        self.label = widgets.Label(
            value='Max cluster diameter:',
            # # layout=self.widget.layout,
            # display='flex',
            # flex_flow='column',
        )

        # Update the number of boxes, and create the initial widget
        self._update_boxes(num_boxes)
        initial_widget = self.make_widget()
        super().__init__(name, initial_widget, priority=priority)
        # The output object we will be drawing the MCD object on
        self.output = widgets.Output()

    def make_widget(self):
        return widgets.VBox(children=self.mcd_boxes)

    def update_widget(self, num_boxes):
        self._update_boxes(num_boxes)
        self.widget = self.make_widget()
        # The output has already been displayed, so it doesn't need to be updated
        self.display_widget(display_output=False)

    @property
    def value(self):
        # Return a NumPy array with the MCD values
        return np.array([child.value for child in self.widget.children])

    def _update_boxes(self, num_boxes):
        def add_box():
            """Helper function for making a bound float box"""
            i = len(self.mcd_boxes)
            box = widgets.BoundedFloatText(
                description=f'{i+2}-body',
                value=self.DEFAULT_MCD_VALUE,
                min=0,
                disabled=False,
            )
            self.mcd_boxes.append(box)

        delta = len(self.mcd_boxes) - num_boxes

        if delta > 0:
            # We have more boxes than MCS, pop the difference
            for _ in range(delta):
                try:
                    self.mcd_boxes.pop()
                except IndexError:
                    # mcd_boxes is empty, no need to continue
                    break
        elif delta < 0:
            # We need to add boxes
            for _ in range(abs(delta)):
                add_box()
        logger.debug('Number of boxes in MCD object: %s', len(self.mcd_boxes))

    def set_value(self, value):
        if not len(value) == len(self.mcd_boxes):
            raise ValueError('Incorrect length of value. Got {}, expected {}',
                             len(value), len(self.mcd_boxes))
        for ii, val in enumerate(value):
            self.mcd_boxes[ii].value = val

    def display_widget(self, display_output=True):
        """Display the widget.

        :param display_output: call display() on the output oject.
            Should only be True on the first call, to avoid repeated objects, unless it needs
            to be redrawn after a clear_output()
        """
        with self.output:
            # We are the only ones drawing in this output,
            # so we always clear on an update
            clear_output(wait=True)
            vbox = widgets.VBox(children=[self.label, self.widget])
            vbox.layout.align_items = 'center'
            display(vbox)
            # display(self.label)
            # display(self.widget)
        if display_output:
            logger.debug('Drawing mcd widget output')
            # We don't always want to redraw the entire output, e.g. when simply redrawing
            # this widget in its own output
            display(self.output)


class MaxClusterWidget(WidgetCollection):
    """Widget for doing max cluster size and max cluster diameter"""
    MAX_CLUSTER_SIZE_NAME = 'max_cluster_size'
    MAX_CLUSTER_DIA_NAME = 'max_cluster_dia'

    def make_widgets(self):
        widget_lst = []

        self._max_cluster_output = widgets.Output()
        # Max cluster size and max cluser diameter
        self.max_cluster_size = widgets.BoundedIntText(
            value=4,
            min=2,
            description='Max cluster size:',
            **self.DEFAULT_STYLE_KWARGS,
        )

        widget_lst.append(
            SingleWidget(self.MAX_CLUSTER_SIZE_NAME,
                         self.max_cluster_size,
                         priority=10))

        self.mcd_widget = MaxClusterDiaWidget(self.MAX_CLUSTER_DIA_NAME,
                                              self.number_of_boxes(),
                                              priority=9)

        widget_lst.append(self.mcd_widget)

        def on_max_cluster_size_change(change):
            if gui.utils.is_value_change(change):
                self.update_mcd_widget()

        self.max_cluster_size.observe(on_max_cluster_size_change)
        return widget_lst

    def update_mcd_widget(self):
        logger.debug('Updating mcd widget')
        self.mcd_widget.update_widget(self.number_of_boxes())

    def number_of_boxes(self):
        """Helper function which calculates how many MCD
        # boxes we should have"""
        return self.max_cluster_size.value - 1

    def display_single_widget(self, widget: SingleWidget):
        # Display the max cluster dia and max cluster size widget together
        if widget.name == self.MAX_CLUSTER_SIZE_NAME:
            # We display both when calling max_cluster_dia
            return
        if widget.name == self.MAX_CLUSTER_DIA_NAME:
            box = widgets.HBox(
                children=[self.max_cluster_size, self.mcd_widget.output])
            box.layout.align_items = 'baseline'
            self.mcd_widget.display_widget(display_output=False)
            with self._max_cluster_output:
                clear_output()
                display(box)
            display(self._max_cluster_output)
            return
        super().display_single_widget(widget)

    def get_max_cluster_size(self):
        return self.max_cluster_size.value

    def set_widget_value(self, key, value):
        if key == 'max_cluster_size':
            self.max_cluster_size.value = value
            self.update_mcd_widget()
        elif key == 'max_cluster_dia':
            self.mcd_widget.set_value(value)
        else:
            super().set_widget_value(key, value)


class CommonSettingsButtons(MaxClusterWidget, WidgetCollection, ABC):
    """Some buttons which are shared between all the CLEASE settings types,
    most notably the max cluster dia and max cluster size"""
    def __init__(self, app_data, *args, **kwargs):
        # Store the FloatBoxes for the max cluster dia array
        self.app_data = app_data
        super().__init__(*args, **kwargs)

    def make_widgets(self):
        widget_lst = super().make_widgets()

        # Basis func type
        basis_func_type = widgets.Dropdown(
            options=[('Polynomial', 'polynomial'),
                     ('Trigonometric', 'trigonometric'),
                     ('Binary Linear', 'binary_linear')],
            description='Basis function:',
            **self.DEFAULT_STYLE_KWARGS,
        )
        widget_lst.append(
            SingleWidget('basis_func_type', basis_func_type, priority=100))

        # include background atoms
        include_background_atoms = widgets.Dropdown(
            options=[('True', True), ('False', False)],
            value=False,
            description='Background atoms:',
            **self.DEFAULT_STYLE_KWARGS,
        )
        widget_lst.append(
            SingleWidget('include_background_atoms',
                         include_background_atoms,
                         priority=5))

        # db_name
        db_name = widgets.Text(
            value='clease.db',
            description='Database name:',
            **self.DEFAULT_STYLE_KWARGS,
        )
        widget_lst.append(SingleWidget('db_name', db_name, priority=1))
        return widget_lst

    def get_widget_value(self, widget: SingleWidget):
        if widget.name == 'db_name':
            # Add the cwd to the database name
            # cwd = self.app_data[gui.AppDataKeys.CWD]
            # return str(cwd / widget.widget.value)
            return widget.widget.value
        return super().get_widget_value(widget)


class CEBulkButtons(CommonSettingsButtons):
    def make_widgets(self):
        widget_list = super().make_widgets()

        crystalstructure = widgets.Dropdown(
            options=[
                ('Simple Cubic', 'sc'),
                ('FCC', 'fcc'),
                ('BCC', 'bcc'),
                ('HCP', 'hcp'),
                ('Diamond', 'diamond'),
                ('Zincblende', 'zincblende'),
                ('Rocksalt', 'rocksalt'),
                ('Cesium Chloride', 'cesiumchloride'),
                ('Fluorite', 'fluorite'),
                ('Wurtzite', 'wurtzite'),
            ],
            description='Crystal Structure:',
            **self.DEFAULT_STYLE_KWARGS,
        )

        a = widgets.BoundedFloatText(
            value=None,
            min=0,
            max=999,
            description='a:',
            description_tooltip=(
                'Setting the value to 0 corresponds to setting it to None'),
            **self.DEFAULT_STYLE_KWARGS,
        )

        c = widgets.BoundedFloatText(
            value=None,
            min=0,
            max=999,
            description='c:',
            description_tooltip=(
                'Setting the value to 0 corresponds to setting it to None'),
            **self.DEFAULT_STYLE_KWARGS,
        )

        covera = widgets.BoundedFloatText(
            value=None,
            min=0,
            max=999,
            description='c/a:',
            description_tooltip=(
                'Setting the value to 0 corresponds to setting it to None'),
            **self.DEFAULT_STYLE_KWARGS,
        )

        widget_list.extend([
            SingleWidget('crystalstructure', crystalstructure, priority=80),
            SingleWidget('a', a, priority=75),
            SingleWidget('c', c, priority=70),
            SingleWidget('covera', covera, priority=65),
        ])
        return widget_list

    def get_widget_value(self, widget: SingleWidget):
        """Helper function for getting the value from a button."""
        if widget.name in ['a', 'c', 'covera']:
            # Set value to 0 to ignore
            value = widget.widget.value
            if value == 0:
                return None
        return super().get_widget_value(widget)


class CECrystalButtons(CommonSettingsButtons):
    def make_widgets(self):
        widget_list = super().make_widgets()
        spacegroup = widgets.IntText(
            value=1,
            description='Spacegroup:',
            **self.DEFAULT_STYLE_KWARGS,
        )

        widget_list.append(SingleWidget('spacegroup', spacegroup, priority=80))
        return widget_list


class CESlabButtons(CommonSettingsButtons):
    def make_widgets(self):
        widget_list = super().make_widgets()

        millers = []
        coord = ['x', 'y', 'z']
        for c in coord:
            millers.append(
                widgets.IntText(
                    value=4,
                    description=f'Miller ({c}):',
                    **self.DEFAULT_STYLE_KWARGS,
                ))
        miller_box = widgets.VBox(children=millers)

        widget_list.append(SingleWidget('miller', miller_box, priority=80))
        return widget_list

    def get_widget_value(self, widget: SingleWidget):
        """Helper function for getting the value from a button."""
        if widget.name == 'miller':
            value = np.zeros(3)
            wdgt = widget.widget
            for ii in range(3):
                value[ii] = wdgt.children[ii].value
            return value
        return super().get_widget_value(widget)


class FixedSizeMode(WidgetCollection):
    def make_widgets(self):
        widgets_list = []
        self.cell_size = widgets.Text(
            description='Cell size:',
            value='3, 3, 3',
            **self.DEFAULT_STYLE_KWARGS,
        )

        label_msg = """Example: "1, 2, 3" for (1, 2, 3) or "3" for (3, 3, 3). """
        label = widgets.Label(value=label_msg)

        box = widgets.HBox(children=[self.cell_size, label])

        widgets_list.append(SingleWidget('size', box))
        return widgets_list

    def get_widget_value(self, widget: SingleWidget):
        if widget.name == 'size':
            cell_widget = widget.widget.children[0]
            return self._parse_text(cell_widget.value)
        return super().get_widget_value(widget)

    def _parse_text(self, text):
        orig_text = str(text)
        text = text.strip()
        text = text.replace(',', ' ')
        parts = text.split()
        if len(parts) == 1:
            rep = int(parts[0])
            return tuple(rep for _ in range(3))
        elif len(parts) == 3:
            return tuple(map(int, parts))
        else:
            raise ValueError(
                'Cannot determine size from string: {}'.format(orig_text))

    def set_widget_value(self, key, value):
        if key == 'size':
            self.cell_size.value = value
        else:
            super().set_widget_value(key, value)


class SupercellFactorSizeMode(WidgetCollection):
    def make_widgets(self):
        widgets_list = []
        supercell_factor = widgets.BoundedIntText(
            description='Supercell Factor:',
            value=27,
            min=0,
            max=999,
            **self.DEFAULT_STYLE_KWARGS)

        widgets_list.append(
            SingleWidget('supercell_factor', supercell_factor, priority=50))

        skew_threshold = widgets.BoundedIntText(
            description='Skew Treshold:',
            value=40,
            min=0,
            max=999,
            **self.DEFAULT_STYLE_KWARGS,
        )
        widgets_list.append(
            SingleWidget('skew_threshold', skew_threshold, priority=50))
        return widgets_list



from abc import ABC, abstractmethod
from typing import Dict, Any
import clease
from clease_gui import WidgetStyleMixin, AppDataKeys

__all__ = ['BaseDashboard']


class BaseDashboard(WidgetStyleMixin, ABC):
    def __init__(self, app_data: Dict[str, Any], initialize=True):
        """Base class for dashboards.

        :param app_data: Dictionary with application data, which will be passed around
            to all dashboards, and possibly widgets if they need them.
        :param initialize: Bool, toggle whether the ``initialize`` function is called.
            Mainly useful for testing purposes. Should generally be set to True.
        """
        # We make app_data a property, as to protect it, so it's not changed into
        # a new object. Can still be mutated.
        self._app_data = app_data
        # Create access to the constant app data keys through ".KEYS.<some-key>"
        self.KEYS = AppDataKeys
        if initialize:
            self.initialize()

    @property
    def app_data(self) -> Dict[str, Any]:
        """Return the app data"""
        return self._app_data

    def get_cwd(self):
        return self.app_data[self.KEYS.CWD]

    @property
    def dev_mode(self) -> bool:
        """Are we in developer mode?"""
        return self.app_data.get(self.KEYS.DEV_MODE, False)

    @property
    def debug(self) -> bool:
        """Alternative name for dev mode"""
        return self.dev_mode

    def log_error(self, logger, *args, **kwargs) -> None:
        """Log as exception if we're in dev mode, otherwise only log as error"""
        if self.dev_mode:
            logger.exception(*args, **kwargs)
        else:
            logger.error(*args, **kwargs)

    @abstractmethod
    def initialize(self):
        """Initialize any widgets related to the dashboard"""

    @abstractmethod
    def display(self):
        """Display the dashboard"""

    @property
    def settings(self) -> clease.settings.ClusterExpansionSettings:
        try:
            return self.app_data[self.KEYS.SETTINGS]
        except KeyError:
            # Raise a better message
            raise KeyError(
                'No settings present. Create/load a settings object first.')

    def get_db_name(self):
        """Return the DB name form the settings, and prepend the current
        working directory"""
        settings = self.settings
        cwd = self.app_data[self.KEYS.CWD]
        return cwd / settings.db_name



from setuptools import setup, find_packages

# Get version number
about = {}
with open('clease_gui/version.py') as f:
    exec(f.read(), about)
version = about['__version__']

EXTRAS_REQUIRE = {
    'dev': (
        'pre-commit',
        'yapf',
        'prospector',
        'pylint',
        'twine',
    ),
    'test': (
        'pytest>=4',
        'pytest-mock',
        'pytest-cases',
    ),
    'doc': ('sphinx', 'sphinx_rtd_theme'),
}

EXTRAS_REQUIRE['full'] = set(value for tup in EXTRAS_REQUIRE.values()
                             for value in tup)

setup(
    name="clease-gui",
    author='Alexander S. Tygesen',
    author_email="alexty@dtu.dk",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://gitlab.com/computationalmaterials/clease-gui',
    version=version,
    description="CLuster Expansion in Atomistic Simulation Environment GUI",
    packages=find_packages(include=['clease_gui', 'clease_gui.*']),
    license='MPL-2.0',
    keywords=[
        'Cluster Expansion',
        'Monte Carlo',
        'Computational materials',
        'Materials research',
        'GUI',
        'Graphical User Interface',
    ],
    project_urls={
        'Documentation': 'https://clease-gui.readthedocs.io/',
        'Source': 'https://gitlab.com/computationalmaterials/clease-gui',
    },
    python_requires=">=3.7",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    install_requires=[
        'ipython',
        'jupyter',
        # Set to master branch for now, since we need some things from there.
        # 'clease @ git+https://gitlab.com//computationalmaterials/clease.git@master',
        'clease>=0.10.4',
        'pandas',
        'numpy',
        'matplotlib',
        'ase',
    ],
    extras_require=EXTRAS_REQUIRE,
    include_package_data=True,  # Include stuff in MANIFEST.in
)
